use strict;
use POSIX;
$|=1;

my $L = 200;
my $bin_size = 10000; # Bin size for doing linear regression
my $bin_sizeM = 1000; # Bin size for doing linear regression for mitochondria
my $max_count = 3;    # Maximum number of reads using the same starting nulceotide
my $extended_region_size = 2000;
my $Pf = 1;           # Fraction of putative regions to exclude from linear regression.

my $eland_filename = "PolII/sgr_files/PolII_eland_result.";
my $input_filename = "Input/sgr_files/Input_eland_result.";
my $bed_filename   = "bed_files/PolII/PolII_sites.";
my $sgr_filename   = "PolII/sgr_files/PolII.";

open OUT, ">PolII.Pf1.txt";

for (my $i=1; $i<=25; $i++) {

    my $chr = "chr";
    if ($i <= 22) {
        $chr .= $i;
    }
    elsif ($i == 23) {
        $chr .= "X";
    }
    elsif ($i == 24) {
        $chr .= "Y";
    }
    elsif ($i == 25) {
        $chr .= "M";
        $bin_size = $bin_sizeM;
    }

    my $infile1 = $input_filename.$chr.".txt";
    my $infile2 = $eland_filename.$chr.".txt";
    my $infile3 = $bed_filename.$chr.".bed";

    # Read in potential target regions

    open IN, "$infile3";
    my %potential_regions = ();
    my $number_potential_regions = 0;

    while (<IN>) {
        chomp;
        my ($chr, $start, $stop) = split /\t/, $_;
        $potential_regions{$start} = $stop;
        $number_potential_regions++;
    }

    close IN;

    # Create a hash with all the reads start wrt forward strand for the ChIP-Seq sample

    open IN, "$infile2";
    my %starts_sample = ();

    while (<IN>) {
        chomp;
        my ($t1, $seq, $map, $t3, $t4, $t5, $chrt, $pos, $str, @rest) = split /\t/, $_;
        my $read_length = length $seq;

        $pos = $pos + $read_length - $L if($str eq "R");
        $starts_sample{$pos} += 1;
        $starts_sample{$pos} = $max_count if ($starts_sample{$pos} > $max_count);
    }
    close IN;

    my @sample_positions = sort {$a <=> $b} keys %starts_sample;   # Sorted list of sample start positions
    my $no_sample_positions = @sample_positions;

    # Create a hash with all the reads start wrt forward strand for the control

    open IN, "$infile1";
    my %starts_control = ();

    while (<IN>) {
        chomp;
        my ($t1, $seq, $map, $t3, $t4, $t5, $chrt, $pos, $str, @rest) = split /\t/, $_;
        my $read_length = length $seq;

        $pos = $pos + $read_length - $L if ($str eq "R");
        $starts_control{$pos} += 1;
        $starts_control{$pos} = $max_count if ($starts_control{$pos} > $max_count);
    }
    close IN;

    my @control_positions = sort {$a <=> $b} keys %starts_control;   # Sorted list of control start positions
    my $no_control_positions = @control_positions;

    # Perform linear regression between Sample and Control overs counts of tags per bin ($bin_size) to get scaling factor between control and sample

    my %x = ();  # bins of counts of control tags
    my %y = ();  # bins of counts of ChIP-Seq tags

    foreach my $pos (sort {$a <=> $b} keys %starts_control) {
        $x{int($pos / $bin_size)} += 1;
    }

    foreach my $pos (sort {$a <=> $b} keys %starts_sample) {
        $y{int($pos / $bin_size)} += 1;
    }

    my ($n, $sx, $sy, $sxx, $sxy) = (0,0,0,0,0);
    foreach my $bin (keys %x) {
        my $bin_start = $bin * $bin_size;
        my $bin_stop  = ($bin + 1) * $bin_size - 1;

        my $flag = 0;
        my $region_count = 1;
        foreach my $region_start (keys %potential_regions) {
            my $region_stop = $potential_regions{$region_start};

            if ($region_count < $Pf * $number_potential_regions) {
                $flag = 1 if ($region_start <= $bin_stop && $bin_start <= $region_stop);
            }
            $region_count++;
        }

        if (exists $y{$bin} && $flag == 0) {
            $n++;
            $sx += $x{$bin};
            $sy += $y{$bin};
            $sxx += $x{$bin} * $x{$bin};
            $sxy += $x{$bin} * $y{$bin};
        }
    }

    my $scaling_factor = ($n * $sxy - $sx * $sy) / ($n * $sxx - $sx * $sx); # This is the scaling factor to normalize control counts to sample counts

    # For each site from the bed file count the number of tags in the regions from both the sample and control

    open IN, "$infile3";

    while (<IN>) {
        chomp;
        my ($tchr, $start, $stop) = split /\t/, $_; # Get site from bed file

        my $tags_sample           = 0;
        my $tags_control          = 0;
        my $tags_control_extended = 0;

        # Count the number of sample tags in the region

        my $index = int ($no_sample_positions / 2);
        my $shift = int ($no_sample_positions / 4);
        my $flag = 0;

        while ($flag == 0) {
            my $pos      = $sample_positions[$index];
            my $prec_pos = $sample_positions[$index - 1];

            if ($prec_pos < $start - $L && $pos >= $start - $L) {
                $flag = 1;
            }
            elsif ($index == 1 || $index == $no_sample_positions - 1) {
                $flag = 1;
            }
            elsif ($pos < $start - $L) {
                $index += $shift;
                $shift = max(1, int ($shift / 2));
            }
            elsif ($prec_pos >= $start - $L) {
                $index -= $shift;
                $shift = max(1, int ($shift / 2));
            }
        }

        while ($sample_positions[$index] <= $stop && $index < $no_sample_positions) {
            $tags_sample += $starts_sample{$sample_positions[$index]};
            $index++;
        }

        # Count the number of control tags in the region

        my $index = int ($no_control_positions / 2);
        my $shift = int ($no_control_positions / 4);
        my $flag = 0;

        while ($flag == 0) {
            my $pos      = $control_positions[$index];
            my $prec_pos = $control_positions[$index - 1];

            if ($prec_pos < $start - $L && $pos >= $start - $L) {
                $flag = 1;
            }
            elsif ($index == 1 || $index == $no_control_positions - 1) {
                $flag = 1;
            }
            elsif ($pos < $start - $L) {
                $index += $shift;
                $shift = max(1, int ($shift / 2));
            }
            elsif ($prec_pos >= $start - $L) {
                $index -= $shift;
                $shift = max(1, int ($shift / 2));
            }
        }

        while ($control_positions[$index] <= $stop && $index < $no_control_positions) {
            $tags_control += $starts_control{$control_positions[$index]};
            $index++;
        }

        # Count the number of control tags in an extended region

        if ($stop - $start <= 1000) {
            my $index = int ($no_control_positions / 2);
            my $shift = int ($no_control_positions / 4);
            my $flag = 0;

            while ($flag == 0) {
                my $pos      = $control_positions[$index];
                my $prec_pos = $control_positions[$index - 1];

                if ($prec_pos < $start - $L - $extended_region_size && $pos >= $start - $L- $extended_region_size) {
                    $flag = 1;
                }
                elsif ($index == 1 || $index == $no_control_positions - 1) {
                    $flag = 1;
                }
                elsif ($pos < $start - $L - $extended_region_size) {
                    $index += $shift;
                    $shift = max(1, int ($shift / 2));
                }
                elsif ($prec_pos >= $start - $L - $extended_region_size) {
                    $index -= $shift;
                    $shift = max(1, int ($shift / 2));
                }
            }

            while ($control_positions[$index] <= $stop + $extended_region_size && $index < $no_control_positions) {
                $tags_control_extended += $starts_control{$control_positions[$index]};
                $index++;
            }
        }

        # Use either the count of control tags from the extended region (normalized) or the actaul regions whichever is bigger

        $tags_control_extended = $tags_control_extended * ($stop - $start + 1) / ($stop - $start + 1 + 2 * $extended_region_size);
        $tags_control = max($tags_control, $tags_control_extended);

        my $adjusted_tags_control = ceil($tags_control * $scaling_factor);
        $adjusted_tags_control = 1 if ($adjusted_tags_control == 0);
        $tags_control = 1 if ($tags_control == 0);

        my $enrich = $tags_sample / ($tags_control * $scaling_factor);
        my $excess = $tags_sample - $adjusted_tags_control;

        my $n = $tags_sample + $adjusted_tags_control;
        my $k = $adjusted_tags_control;
        my $p = 0.5;

        my $pval = 0;
        for (my $a = 0; $a <= $k; $a++) {
            $pval += choose($n, $a) * $p**$a * (1 - $p)**($n - $a);
        }
        if (($pval eq "nan" || $pval eq "inf") && $k < $n - $k) {
            $pval = 0;
        }
        elsif (($pval eq "nan" || $pval eq "inf") && $k > $n - $k) {
            $pval = 1;
        }
        print OUT "$chr\t$start\t$stop\t$tags_sample\t$adjusted_tags_control\t$enrich\t$excess\t$pval\n";
    }
    close IN;
    print "$chr\n";
}

close OUT;

sub choose {
    my ($n, $i) = @_;

    my $res = 1;
    if ($i <= $n - $i) {
        for (my $j = 0; $j < $i; $j++) {
            $res = $res * ($n - $j) / ($i - $j);
        }
    }
    else {
        for (my $j = 0; $j < $n - $i; $j++) {
            $res = $res * ($n - $j) / ($n - $i - $j);
        }
    }
    return $res;
}

sub max {
    my ($a, $b) = @_;
    if ($a > $b) {
        return $a;
    }
 else {
        return $b;
    }
}
