use strict;
$|=1;

my $L = 200;               # Average length of DNA fragments
my $window_size = 1000000; # Window size for scoring
my $max_threshold = 100;   # Maximum allowed threshold
my $max_gap = 200;         # The maximum gap allowed between hits that are merged together - hits that have greater seperations arent merged
my $FDR_required = 0.05;   # Required false discovery rate
my $number_of_sims = 10;   # Number of simulations per window to estimate FDR

my $eland_filename = "PolII/sgr_files/PolII_eland_result.";
my $sgr_filename   = "PolII/sgr_files/PolII.";
my $bed_filename   = "bed_files/PolII/PolII_sites.";
my $map_filename   = "Code/Mapability_HG.txt";

my %mapability_fraction = ();

open IN, "$map_filename";
while (<IN>) {
    chomp;
    my ($chr, $window, $count) = split /\s+/, $_;
    $ {$mapability_fraction{$chr}}{$window} = $count / $window_size if ($count > 0);
}
close IN;

for (my $i=1; $i<=25; $i++) {

  my $chr = "chr";

  if ($i <= 22) {
    $chr .= $i;
  }
  elsif ($i == 23) {
    $chr .= "X";
  }
  elsif ($i == 24) {
    $chr .= "Y";
  }
  elsif ($i == 25) {
    $chr .= "M";
}

# Count the number of reads per window using the eland_results file

  my %reads_per_window = ();
  my $total_number_of_reads = 0;

  my $filename = $eland_filename.$chr.".txt";
  open IN, "$filename";
  while (<IN>) {
    chomp;
    my ($t1, $seq, $map, $t2, $t3, $t4, $t5, $pos, @rest) = split /\t/, $_;
    $reads_per_window{int($pos / $window_size)} += 1;

    $total_number_of_reads++;
  }
  close IN;
  my @list_of_sorted_windows = sort {$a <=> $b} keys %reads_per_window;
  my $highest_window = @list_of_sorted_windows[$#list_of_sorted_windows];

# Load the sgr file into memory

  my %map = ();       # Map from sgr file
  my @positions = (); # Sorted positions from sgr files along chromosome

  my $filename = $sgr_filename.$chr.".sgr";
  open IN, "$filename";
  while (<IN>) {
      chomp;
      my ($temp_chr, $pos, $height) = split /\t/, $_;
      push @positions, $pos;
      $map{$pos} = $height;
  }
  close IN;

# For each possible threshold compute all the peaks

  my %merged_actual_peaks = ();

  for (my $threshold = 2; $threshold <= $max_threshold; $threshold++) {
      my ($flag, $start, $stop, $height) = (0, 0, 0, 0);
      my @actual_peaks = ();

      foreach my $pos (@positions) {
	  my $height = $map{$pos};
	  if ($flag == 0 && $height >= $threshold) {
	      $start = $pos;
	      $flag = 1;
	  }
	  elsif ($flag == 1 && $height < $threshold) {
	      $stop = $pos - 1;
	      push @actual_peaks, "$start\t$stop";
	      $flag = 0;
	  }
      }
      
      # Merge peaks/sites that are seperated by less than 200 bp

      my ($peak_start, $peak_stop) = (-1000, -1000); 
      
      foreach (@actual_peaks) {
	  my ($start, $stop) = split /\t/, $_;
	  
	  if ($start - $peak_stop <= $max_gap) {
	      $peak_stop = $stop;
	  }
	  elsif ($start - $peak_stop > $max_gap) {
	      if ($peak_start >= 0) {
		  push @ {$merged_actual_peaks{$threshold}}, "$peak_start\t$peak_stop";
	      }
	      $peak_start = $start;
	      $peak_stop = $stop;
	  }
      }
      if ($peak_start >= 0) {
	  push @ {$merged_actual_peaks{$threshold}}, "$peak_start\t$peak_stop";
      }  
  }


# Analyze each window seperately
  
  my @list_of_sites = ();

  for (my $window = 0; $window <= $highest_window; $window++) {
       if ($reads_per_window{$window} > 0) {
	  
	  my ($window_start, $window_stop) = ($window * $window_size + 1, ($window + 1) * $window_size);
	  
	  # Generate simulated maps of random peaks for the same window with the same number of fragments

	  my %averaged_simulated_counts = ();
	  foreach (my $sim=1; $sim<= $number_of_sims; $sim++) {
	  
	      my %simulated_data = ();
	      for (my $j=1; $j<= $reads_per_window{$window}; $j++) {
		  my $pos = int(rand($window_size * $ {$mapability_fraction{$chr}}{$window}));
		  $simulated_data{$pos} += 1;
		  $simulated_data{$pos + $L} -= 1;
	      } 
	      my @sorted_simulated_positions = sort {$a <=> $b} (keys %simulated_data);

	      for (my $threshold = 2; $threshold <= $max_threshold; $threshold++) {
		  
		  my $flag   = 0;
		  my $start  = 0;
		  my $stop   = 0;
		  my $height = 0;
		  my @simulated_sites = ();
		  
		  foreach my $pos (@sorted_simulated_positions) {
		      $height += $simulated_data{$pos};
		      
		      if ($flag == 0 && $height >= $threshold) {
			  $start = $pos;
			  $flag = 1;
		      }

		      elsif ($flag == 1 && $height < $threshold) {
			  $stop = $pos;
			  
			  $start = 0 if ($start < 0);
			  my $line =  "$start\t$stop";
			  push @simulated_sites, $line; 
			  
			  $flag = 0;
		      }
		  }
		  
		  my $count_simulated_sites = 0;
		  my ($peak_start, $peak_stop) = (-1000, -1000); 
      
		  foreach (@simulated_sites) {
		      my ($start, $stop) = split /\t/, $_;
	  
		      if ($start - $peak_stop <= $max_gap) {
			  $peak_stop = $stop;
		      }
		      elsif ($start - $peak_stop > $max_gap) {
			  $count_simulated_sites++ if ($peak_start >= 0);
			  $peak_start = $start;
			  $peak_stop = $stop;
		      }
		  }
		  $count_simulated_sites++ if ($peak_start >= 0);
		  $averaged_simulated_counts{$threshold} += $count_simulated_sites / $number_of_sims;
	      }
	  }	  

	  # Pick the correct threshold for the window

	  my $FDR          = 1.0;
	  my $previous_FDR = 1.0;
	  my $count_actual_peaks_in_window = 0;
	  my $threshold = 2;
	  while ($threshold <= $max_threshold && ($previous_FDR > $FDR_required || $previous_FDR < $FDR)) {
	      
	      # Count the number of actual peaks in the window above the threshold

	      $count_actual_peaks_in_window = 0;
	      foreach (@ {$merged_actual_peaks{$threshold}}) {
		  my ($peak_start, $peak_stop) = split /\t/, $_;
		  $count_actual_peaks_in_window++ if($peak_start <= $window_stop && $window_start <= $peak_stop);
	      }

	      $previous_FDR = $FDR;
	      $FDR = 0;
	      $FDR = $averaged_simulated_counts{$threshold} / $count_actual_peaks_in_window if ($count_actual_peaks_in_window > 0);
	      
	      $threshold++;
	  }
	  $threshold -= 2;
	  $FDR = $previous_FDR;

	  foreach (@ {$merged_actual_peaks{$threshold}}) {
	      my ($peak_start, $peak_stop) = split /\t/, $_;
	      if($peak_start <= $window_stop && $window_start <= $peak_stop) {
		  $peak_start = $window_start if ($window_start >= $peak_start);
		  $peak_stop  = $window_stop  if ($window_stop <= $peak_stop );
		  push @list_of_sites, "$peak_start\t$peak_stop";
	      }
	  }
	  print "$chr\t$window\t$reads_per_window{$window}\t$threshold\t$FDR\t$count_actual_peaks_in_window\n";
	  
      }
  }

  my $filename = $bed_filename.$chr.".bed";
  open OUT, ">$filename";

  my ($peak_start, $peak_stop) = (-1000, -1000); 
      
  foreach (@list_of_sites) {
      my ($start, $stop) = split /\t/, $_;
      
      if ($start - $peak_stop <= $max_gap) {
	  $peak_stop = $stop;
      }
      elsif ($start - $peak_stop > $max_gap) {
	  if ($peak_start >= 0) {
	      print OUT "$chr\t$peak_start\t$peak_stop\n";
	  }
	  $peak_start = $start;
	  $peak_stop = $stop;
      }
  }
  if ($peak_start >= 0) {
      print OUT "$chr\t$peak_start\t$peak_stop\n";
  }  
  close OUT;

}
